;; Generated package description from js2-mode.el  -*- no-byte-compile: t -*-
(define-package "js2-mode" "20231224" "Improved JavaScript editing mode" '((emacs "24.1") (cl-lib "0.5")) :commit "010a536da6df345a0b9a55bbd892f5f4d0d9fdfc" :authors '(("Steve Yegge" . "steve.yegge@gmail.com") ("mooz" . "stillpedant@gmail.com") ("Dmitry Gutov" . "dgutov@yandex.ru")) :maintainer '(("Steve Yegge" . "steve.yegge@gmail.com") ("mooz" . "stillpedant@gmail.com") ("Dmitry Gutov" . "dgutov@yandex.ru")) :keywords '("languages" "javascript") :url "https://github.com/mooz/js2-mode/")
